#include "memmanager.h"
#include "system.h"

MemoryManager::MemoryManager()
{
	swapBitMap=new BitMap(SWAPSIZE);
	fileSystem->Create(SWAPFILE,0);
	swapFile=fileSystem->Open(SWAPFILE);

	char *tmp=new char[SWAPSIZE*PageSize];
	swapFile->Write(tmp,SWAPSIZE*PageSize);
	delete[] tmp;
	delete swapFile;
	swapFile=fileSystem->Open(SWAPFILE);

	physicalPage=new TranslationEntry*[NumPhysPages];

	for(int i=0;i<NumPhysPages;i++)
		physicalPage[i]=NULL;
}

MemoryManager::~MemoryManager()
{
	delete swapBitMap;
	delete swapFile;
	delete[] physicalPage;
}

int MemoryManager::getPageFaultPos()
{
	return machine->ReadRegister(BadVAddrReg)/PageSize;
}

int MemoryManager::getKillTLB()
{
	for(int i=0;i<TLBSize;i++)
		if(machine->tlb[i].valid==FALSE)
			return i;
	int killTLB;

	while(1)
	{
		if(machine->tlb[TLBPos].use==TRUE)
			machine->tlb[TLBPos].use=FALSE;
		else
		{
			if(machine->tlb[TLBPos].dirty==TRUE)
				physicalPage[machine->tlb[TLBPos].physicalPage]->dirty=TRUE;
			killTLB=TLBPos;
			TLBPos=(TLBPos+1)%TLBSize;
			break;
		}
		TLBPos=(TLBPos+1)%TLBSize;
	}
	return killTLB;
}

int MemoryManager::getKillPage()
{
	if(machine->bitmap->NumClear()>0)
		return machine->bitmap->Find();
	int killPage;
	while(1)
	{
		if(physicalPage[pagePos]->use==TRUE)
			physicalPage[pagePos]->use=FALSE;
		else
		{
			for(int i=0;i<TLBSize;i++)
			{
				if(machine->tlb[i].valid==TRUE&&physicalPage[pagePos]->physicalPage==machine->tlb[i].physicalPage)
				{
					if(machine->tlb[i].dirty==TRUE)
						physicalPage[pagePos]->dirty=TRUE;
					machine->tlb[i].valid=FALSE;
				}
			}

			if(physicalPage[pagePos]->dirty==TRUE)
			{
				swapFile->WriteAt(machine->mainMemory+physicalPage[pagePos]->physicalPage*PageSize,PageSize,physicalPage[pagePos]->swapPos*PageSize);
				physicalPage[pagePos]->dirty=FALSE;
			}
			physicalPage[pagePos]->valid=FALSE;
			killPage=pagePos;
			pagePos=(pagePos+1)%NumPhysPages;
			break;
		}
		pagePos=(pagePos+1)%NumPhysPages;
	}
	return killPage;
}

void MemoryManager::PageFault()
{
//	printf("\nPageFault begin\n");
	int faultPage=getPageFaultPos();
//	printf("faultpage %d\n",faultPage);
	TranslationEntry *pageTable=currentThread->space->pageTable;

	if(pageTable[faultPage].valid) // page table hit
	{
//		printf("page table hit %d\n",faultPage);
		int killTLB=getKillTLB();
//		printf("kill tlb %d\n",killTLB);
		machine->tlb[killTLB].virtualPage=faultPage;
		machine->tlb[killTLB].physicalPage=pageTable[faultPage].physicalPage;
		machine->tlb[killTLB].swapPos=pageTable[faultPage].swapPos;
		machine->tlb[killTLB].valid=TRUE;
		machine->tlb[killTLB].use=TRUE;
		machine->tlb[killTLB].readOnly=FALSE;
		machine->tlb[killTLB].dirty=FALSE;
	}
	else // page table miss
	{
//		printf("page table miss %d\n",faultPage);
		int killPage=getKillPage();
//		printf("kill page %d\n",killPage);
		readDisk(machine->mainMemory+killPage*PageSize,pageTable[faultPage].swapPos);
		pageTable[faultPage].physicalPage=killPage;
		pageTable[faultPage].valid=TRUE;
		pageTable[faultPage].use=TRUE;
		pageTable[faultPage].readOnly=FALSE;
		pageTable[faultPage].dirty=FALSE;

		physicalPage[killPage]=&pageTable[faultPage];

	}
//	dump();
//	printf("PageFault end\n");
}

int MemoryManager::writeDisk(char *buff,int pos,int len)
{
	int writeSize;
	writeSize=swapFile->WriteAt(buff,len,pos*PageSize);
	return writeSize;
}

int MemoryManager::readDisk(char *buff,int pos)
{
	int readSize;
	readSize=swapFile->ReadAt(buff,PageSize,pos*PageSize);
	return readSize;
}

int MemoryManager::find()
{
	return swapBitMap->Find();
}

int MemoryManager::numClear()
{
	return swapBitMap->NumClear();
}

void MemoryManager::clear(int pos)
{
	swapBitMap->Clear(pos);
}

void MemoryManager::clearPage(int page)
{
	physicalPage[page]=NULL;
}

void MemoryManager::clearTLB()
{
	for(int i=0;i<TLBSize;i++)
	{
		if(machine->tlb[i].dirty&&machine->tlb[i].valid)
		{
			if(physicalPage[machine->tlb[i].physicalPage]!=NULL)
				physicalPage[machine->tlb[i].physicalPage]->dirty=TRUE;
		}
	}
}

void MemoryManager::dump()
{
	printf("thread : %d\n",(int)currentThread);
	printf("tlb\tphys\tvirt\tswap\tvurd\n");
	for(int i=0;i<TLBSize;i++)
		printf("%d\t%d\t%d\t%d\t%d%d%d%d\n",i,machine->tlb[i].physicalPage,machine->tlb[i].virtualPage,machine->tlb[i].swapPos,machine->tlb[i].valid,machine->tlb[i].use,machine->tlb[i].readOnly,machine->tlb[i].dirty);

/*	printf("phys\tphys\tvirt\tswap\tvurd\n");
	for(int k=0;k<NumPhysPages;k++)
	{
		if(physicalPage[k]!=NULL)
			printf("%d\t%d\t%d\t%d\t%d%d%d%d\n",k,physicalPage[k]->physicalPage,physicalPage[k]->virtualPage,physicalPage[k]->swapPos,physicalPage[k]->valid,physicalPage[k]->use,physicalPage[k]->readOnly,physicalPage[k]->dirty);
	}*/

	TranslationEntry *pageTable=currentThread->space->pageTable;
	int numPages=currentThread->space->numPages;
	printf("page\tphys\tvirt\tswap\tvurd\n");
	for(int j=0;j<numPages;j++)
		printf("%d\t%d\t%d\t%d\t%d%d%d%d\n",j,pageTable[j].physicalPage,pageTable[j].virtualPage,pageTable[j].swapPos,pageTable[j].valid,pageTable[j].use,pageTable[j].readOnly,pageTable[j].dirty);
}
